from __future__ import annotations

import sys
import typing as t
from sqlalchemy import Boolean, Column, exists, func
from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import InstrumentedAttribute, joinedload, selectinload
from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.sqltypes import NullType

from aerie.base import Base
from aerie.collections import Collection
from aerie.paginator import Page
from aerie.utils import colorize, convert_exceptions

M = t.TypeVar('M', bound=Base)
E = t.TypeVar('E', bound=Base)


class SelectQuery(t.Generic[M]):
    def __init__(
        self,
        model: t.Type[M],
        executor: AsyncSession,
        base_stmt: Select = None,
        returns_model: bool = None,
    ) -> None:
        self._model: t.Type[Base] = model
        self._executor = executor
        self._stmt: Select = select(model) if base_stmt is None else base_stmt
        self._is_session = isinstance(executor, AsyncSession)
        self._returns_model = self._is_session if returns_model is None else returns_model

    def where(self, *conditions: ColumnElement[Boolean]) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.where(*conditions))

    def having(self, conditions: ColumnElement[Boolean]) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.having(conditions))

    def join(self, target: t.Type[E], on_clause: t.Any = None, full: bool = False) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.join(target, onclause=on_clause, full=full))

    def left_join(self, target: t.Type[E], on_clause: t.Any = None, full: bool = False) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.outerjoin(target, on_clause, full=full))

    def filter_by(self, **kwargs: t.Any) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.filter_by(**kwargs))

    def group_by(self, *clauses: Column) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.group_by(*clauses))

    def order_by(self, *clauses: ColumnElement[NullType]) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.order_by(*clauses))

    def limit(self, limit: int, offset: int = None) -> SelectQuery[M]:
        stmt = self._clone(base_stmt=self._stmt.limit(limit))
        if offset:
            stmt = stmt.offset(offset)
        return stmt

    def offset(self, offset: int) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.offset(offset))

    def slice(self, start: int, stop: int) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.slice(start, stop))

    def for_update(self) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.with_for_update(nowait=True))

    # def cte(self, name, recursive, nesting) -> SelectQuery:
    #     return self._clone(base_stmt=self._stmt.cte(name, recursive, nesting))

    # def subquery(self, name: str) -> SelectQuery:
    #     return self._clone(base_stmt=self._stmt.subquery(name))

    # def scalar_subquery(self) -> SelectQuery:
    #     return self._clone(base_stmt=self._stmt.scalar_subquery())

    def options(self, *options: t.Any) -> SelectQuery[M]:
        return self._clone(base_stmt=self._stmt.options(*options))

    def preload(self, *cols: InstrumentedAttribute) -> SelectQuery[M]:
        return self.options(*[joinedload(col) for col in cols])

    def prefetch(self, *cols: InstrumentedAttribute) -> SelectQuery[M]:
        return self.options(*[selectinload(col) for col in cols])

    def params(self, *args: t.Any, **kwargs: t.Any) -> SelectQuery[M]:
        """Add values for bind parameters which may have been specified in filter()."""
        return self._clone(base_stmt=self._stmt.params(*args, **kwargs))

    # def difference(self, other: SelectQuery) -> SelectQuery:
    #     self._stmt = self._stmt.except_(other)
    #     return self

    # def difference_all(self, other: SelectQuery):
    #     self._stmt = self._stmt.except_all(other)
    #     return self

    # def intersect(self, other: SelectQuery) -> SelectQuery:
    #     self._stmt = self._stmt.intersect(other)
    #     return self

    # def intersect_all(self, other: SelectQuery) -> SelectQuery:
    #     self._stmt = self._stmt.intersect_all(other)
    #     return self

    # def union(self, *q: SelectQuery) -> SelectQuery:
    #     self._stmt = self._stmt.union(*q)
    #     return self

    # def union_all(self, *q: SelectQuery) -> SelectQuery:
    #     self._stmt = self._stmt.union_all(*q)
    #     return self

    def dump(self, writer: t.IO[str] = sys.stdout, colorize_sql: bool = True) -> SelectQuery[M]:
        sql = str(self)
        if colorize_sql:
            sql = colorize(sql)
        writer.write(sql)
        return self

    def to_string(self) -> str:
        return str(self._stmt.compile(dialect=self._executor.bind.dialect, compile_kwargs={"literal_binds": True}))

    async def one(self) -> M:
        with convert_exceptions():
            result = await self._execute(self._stmt)
            return result.scalars().one() if self._returns_model else result.one()

    async def one_or_none(self) -> t.Optional[M]:
        with convert_exceptions():
            result = await self._execute(self._stmt)
            return result.scalars().one_or_none()

    async def first(self) -> t.Optional[M]:
        result = await self._execute(self._stmt.limit(1))
        return result.scalars().first()

    async def all(self) -> Collection[M]:
        result = await self._execute(self._stmt)
        return Collection(result.scalars().all())

    async def exists(self, params: t.Mapping = None) -> bool:
        stmt = select(exists(self._stmt))
        result = await self._execute(stmt, params)
        return result.scalar() is True

    async def count(self, params: t.Mapping = None) -> int:
        stmt = select(func.count('*')).select_from(self._stmt)
        result = await self._execute(stmt, params)
        count = result.scalar()
        return int(count) if count else 0

    async def paginate(self, page: int = 1, page_size: int = 50, params: t.Mapping = None) -> Page:
        offset = (page - 1) * page_size
        total = await self.count(params)
        rows = await self.limit(page_size).offset(offset).all()
        return Page(list(rows), total, page, page_size)

    async def execute(self) -> Result:
        return await self._execute(self._stmt)

    async def _execute(self, stmt: Select, params: t.Mapping = None) -> Result:
        return await self._executor.execute(stmt, params)

    def _clone(self, *, base_stmt: Select, returns_model: bool = None) -> SelectQuery:
        return SelectQuery(
            model=self._model,
            executor=self._executor,
            base_stmt=base_stmt,
            returns_model=returns_model,
        )

    def __await__(self) -> t.Generator[t.Any, None, Collection[M]]:
        return self.all().__await__()

    def __str__(self) -> str:
        return self.to_string()



from sqlalchemy import exc


class AerieError(Exception):  # pragma: no cover
    """Base class for aerie errors."""


class TooManyResultsError(AerieError, exc.MultipleResultsFound):  # pragma: no cover
    """Raised when .one() query matches more that one result."""


class NoResultsError(AerieError, exc.NoResultFound):  # pragma: no cover
    """Raised when .one() matches no rows."""


class NoActiveSessionError(AerieError):  # pragma: no cover
    """Raised when not global session exists."""



import sqlalchemy as sa
import typing as t

from aerie.base import Base
from aerie.collections import Collection
from aerie.queries import SelectQuery
from aerie.session import get_current_session

C = t.TypeVar('C', bound=Base)


class AutoIntegerId(Base):
    __abstract__ = True
    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)


class AutoBigIntegerId(Base):
    __abstract__ = True
    id = sa.Column(sa.BigInteger, primary_key=True, autoincrement=True)


class Queryable(Base):
    __abstract__ = True

    @classmethod
    def query(cls: t.Type[C]) -> SelectQuery[C]:
        return get_current_session().query(cls)

    @classmethod
    async def all(cls: t.Type[C]) -> Collection[C]:
        return await get_current_session().query(cls).all()

    @classmethod
    async def get(cls: t.Type[C], pk: t.Any, pk_column: str = 'id') -> C:
        column = getattr(cls, pk_column)
        return await get_current_session().query(cls).where(column == pk).one()

    @classmethod
    async def get_or_none(cls: t.Type[C], pk: t.Any, pk_column: str = 'id') -> t.Optional[C]:
        column = getattr(cls, pk_column)
        return await get_current_session().query(cls).where(column == pk).one_or_none()

    @classmethod
    async def create(cls: t.Type[C], **values: t.Any) -> C:
        instance = cls(**values)  # type: ignore
        await instance.save()  # type: ignore
        return instance

    async def save(self, commit: bool = True) -> None:
        session = get_current_session()
        session.add(self)  # type: ignore
        if commit:
            await session.commit()

    async def delete(self, commit: bool = True) -> None:
        session = get_current_session()
        await session.delete(self)
        if commit:
            await session.commit()

    async def refresh(self) -> None:
        session = get_current_session()
        await session.refresh(self)



import sqlalchemy as sa
import typing as t
from sqlalchemy.ext.asyncio import AsyncEngine


class Schema:
    def __init__(self, engine: AsyncEngine, metadata: sa.MetaData) -> None:
        self.engine = engine
        self.metadata = metadata

    async def create_tables(self, tables: t.List[t.Union[str, sa.Table]] = None) -> None:
        if tables:
            tables = [self.metadata.tables[table] if isinstance(table, str) else table for table in tables]
        async with self.engine.begin() as connection:
            await connection.run_sync(self.metadata.create_all, tables=tables)

    async def drop_tables(self, tables: t.List[t.Union[str, sa.Table]] = None) -> None:
        if tables:
            tables = [self.metadata.tables[table] if isinstance(table, str) else table for table in tables]

        async with self.engine.begin() as connection:
            await connection.run_sync(self.metadata.drop_all, tables=tables)



import typing as t
from contextlib import contextmanager
from sqlalchemy.exc import MultipleResultsFound, NoResultFound

from aerie.exceptions import NoResultsError, TooManyResultsError


@contextmanager
def convert_exceptions() -> t.Generator[None, None, None]:
    try:
        yield
    except MultipleResultsFound as exc:
        raise TooManyResultsError() from exc
    except NoResultFound:
        raise NoResultsError('No rows found when one was required.')


ITEM = t.TypeVar('ITEM')


def chunked(items: t.Iterable[ITEM], size: int) -> t.Generator[t.List[ITEM], None, None]:
    result = []
    for value in items:
        result.append(value)
        if len(result) == size:
            yield result
            result = []

    if len(result):
        yield result


def colorize(sql: str) -> str:
    try:
        import pygments
        import pygments.formatters
        import pygments.lexers

        lexer = pygments.lexers.get_lexer_by_name("sql")
        formatter = pygments.formatters.get_formatter_by_name("console")
        sql = pygments.highlight(sql, lexer, formatter)
    except ImportError:
        pass
    return sql



from .base import Base, metadata
from .database import Aerie
from .exceptions import AerieError, NoResultsError, TooManyResultsError
from .paginator import Page
from .session import DbSession

__all__ = [
    'Aerie',
    'DbSession',
    'TooManyResultsError',
    'NoResultsError',
    'AerieError',
    'Page',
    'metadata',
    'Base',
]



from __future__ import annotations

import contextvars as cv
import typing as t
from sqlalchemy.ext.asyncio import AsyncSession

from aerie.base import Base
from aerie.exceptions import NoActiveSessionError
from aerie.queries import SelectQuery

M = t.TypeVar('M', bound=Base)


class DbSession(AsyncSession):
    current_session_stack: cv.ContextVar[list[DbSession]] = cv.ContextVar('current_session_stack', default=[])
    """A list of DbSession instances associated with current task."""

    def query(self, model: t.Type[M]) -> SelectQuery[M]:
        return SelectQuery(model, self)

    @classmethod
    def get_current_session(cls) -> DbSession:
        """Return the last activated session associated with current asyncio task."""
        try:
            stack = cls.current_session_stack.get()
            return stack[len(stack) - 1]
        except IndexError:
            raise NoActiveSessionError()

    async def close(self) -> None:
        await super().close()
        DbSession.current_session_stack.get().remove(self)

    @classmethod
    async def close_all(cls) -> None:
        DbSession.current_session_stack.get().clear()
        await super().close_all()

    async def __aenter__(self) -> DbSession:
        DbSession.current_session_stack.get().append(self)
        return self


def get_current_session() -> DbSession:
    """Return an instance of DbSession that is bound to current task."""
    return DbSession.get_current_session()



from __future__ import annotations

import functools
import itertools
import statistics
from typing import Any, Callable, Dict, Generator, Generic, Iterable, List, Optional, Protocol, TypeVar, Union, overload

from aerie.utils import chunked

E = TypeVar("E", bound=object)


class CastsToString(Protocol):  # pragma: nocover
    def __str__(self) -> str:
        ...


StrLike = Union[str, CastsToString]


class EmptyError(ValueError):
    pass


def attribute_reader(obj: Any, attr: str, default: Any = None) -> Any:
    if hasattr(obj, "__getitem__"):
        return obj.get(attr, default)
    return getattr(obj, attr, default)


class Collection(Generic[E], Iterable[E]):
    def __init__(self, items: Iterable[E]) -> None:
        self.items = list(items)
        self._position = 0

    def first(self) -> Optional[E]:
        """Return the first item from the collection."""
        try:
            return next(self)
        except StopIteration:
            return None

    def last(self) -> Optional[E]:
        """Return the last item from the collection."""
        return self.reverse().first()

    def find(self, fn: Callable[[E], Optional[bool]]) -> Optional[E]:
        return self.filter(fn).first()

    def reverse(self) -> Collection[E]:
        """Reverse collection."""
        return Collection(list(reversed(self)))

    def chunk(self, batch: int) -> Generator[List[E], None, None]:
        """Split collection into chunks."""
        return chunked(self, batch)

    def pluck(self, key: StrLike) -> Collection[Any]:
        """Take a attribute/key named `key` from every item
        and return them in a new collection."""
        key = str(key)
        return Collection([attribute_reader(item, key, None) for item in self])

    def avg(self, field: StrLike = None) -> float:
        """Compute average value of all items.
        If `field` is given and items are not mappings/classes
         the exception will be raised."""
        if not len(self):
            raise EmptyError("Cannot find avg value: collection is empty.")
        items = self.pluck(field) if field else self.items
        return statistics.mean(items or [0])

    def min(self, field: StrLike = None) -> float:
        """Find the smallest item.

        When `field` is given it returns the smallest item by `field` value.
        If collection is empty it will raise ArithmeticError."""
        if not len(self):
            raise EmptyError("Cannot find min value: collection is empty.")
        items = self.pluck(field) if field else self.items
        return min(items)

    def max(self, field: StrLike = None) -> float:
        """Find the biggest item.

        When `field` is given it returns the biggest item by `field` value.
        If collection is empty it will raise ArithmeticError."""
        items = self.pluck(field) if field else self.items
        if not len(self):
            raise EmptyError("Cannot find max value: collection is empty.")
        return max(items)

    def sum(self, field: StrLike = None) -> float:
        """Find a sum of all items in a collection.

        If collection is empty it will raise ArithmeticError."""
        items = self.pluck(field) if field else self.items
        if not len(self):
            raise EmptyError("Cannot find sum value: collection is empty.")
        return sum(items)

    def map(self, fn: Callable[[E], Any]) -> Collection:
        """Apply a function on each collection item
        and return a new collection."""
        return Collection(list(map(fn, self)))

    def each(self, fn: Callable[[E, int], None]) -> Collection[E]:
        """Apply a function on each collection item
        and return same collection.

        Callback will receive current item as the first argument
        and current iteration index as the second argument."""
        for index, item in enumerate(self.items):
            fn(item, index)
        return self

    def every(self, fn: Callable[[E], bool]) -> bool:
        """Test if all items match the condition specified by `fn`."""
        return all(map(fn, self))

    def some(self, fn: Callable[[E], bool]) -> bool:
        """Test if at least one item matches the condition specified by `fn`."""
        return any(map(fn, self))

    def prepend(self, item: E) -> Collection[E]:
        """Add an item to the top of collection."""
        return Collection([item, *self])

    def append(self, item: E) -> Collection[E]:
        """Add an item to the bottom of collection."""
        return Collection([*self, item])

    def filter(self, fn: Callable[[E], Optional[bool]]) -> Collection[E]:
        """Filter collection items with `fn`."""
        return Collection(list(filter(fn, self.items)))

    def reduce(self, fn: Callable[[Any, Any], Any], start: Any = None) -> Any:
        return functools.reduce(fn, self.items, start)

    def sort(self, key: Union[Callable, str] = None, reverse: bool = False) -> Collection[E]:
        """Return a new sorted collection from the items in this collection."""
        if isinstance(key, str):
            key = functools.partial(attribute_reader, attr=key)
        # fixme:
        return Collection(list(sorted(self, key=key, reverse=reverse)))  # type: ignore

    def group_by(self, key: Union[Callable[[Any], str], str]) -> Dict[str, List[E]]:
        if isinstance(key, str):
            key = functools.partial(attribute_reader, attr=key)

        groups = itertools.groupby(sorted(self.items, key=key), key=key)
        return {k: list(v) for k, v in groups}

    def key_value(self, key: Union[Callable[[Any], str], str]) -> Dict[Any, E]:
        if isinstance(key, str):
            key = functools.partial(attribute_reader, attr=key)

        return {key(item): item for item in self}

    def as_list(self) -> List[E]:
        return list(self)

    def choices(self, label_col: str = 'name', value_col: str = 'id') -> list[tuple[Any, Any]]:
        return [(attribute_reader(item, value_col), attribute_reader(item, label_col)) for item in self]

    def choices_dict(
        self, label_col: str = 'name', value_col: str = 'id', label_key: str = 'label', value_key: str = 'value'
    ) -> list[dict[Any, Any]]:
        return [
            {value_key: attribute_reader(item, value_col), label_key: attribute_reader(item, label_col)}
            for item in self
        ]

    @overload
    def __getitem__(self, index: slice) -> list[E]:  # pragma: nocover
        ...

    @overload
    def __getitem__(self, index: int) -> E:  # pragma: nocover
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[E, list[E]]:
        if isinstance(index, slice):
            return [item for item in self.items[index.start : index.stop : index.step]]
        return self.items[index]

    def __setitem__(self, key: int, value: E) -> None:
        self.items.insert(key, value)

    def __delitem__(self, key: int) -> None:
        self.items.pop(key)

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Collection[E]:
        return self

    def __next__(self) -> E:
        if self._position < len(self):
            entity = self.items[self._position]
            self._position += 1
            return entity
        raise StopIteration()

    def __contains__(self, item: E) -> bool:
        return item in self.items

    def __eq__(self, other: Any) -> bool:
        return self.items == other.items

    def __reversed__(self) -> Collection[E]:
        return Collection(reversed(list(self)))

    def __str__(self) -> str:
        truncate = 10
        remainder = len(self) - truncate if len(self) > truncate else 0
        contents = ",".join(map(str, self[0:10]))
        suffix = f" and {remainder} items more" if remainder else ""
        return f"<Collection: [{contents}{suffix}]>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aerie']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy[asyncio]>=1.4.27,<2.0.0']

extras_require = \
{'full': ['aiosqlite>=0.17.0,<0.18.0', 'asyncpg>=0.25.0,<0.26.0'],
 'postgresql': ['asyncpg>=0.25.0,<0.26.0'],
 'sqlite': ['aiosqlite>=0.17.0,<0.18.0']}

setup_kwargs = {
    'name': 'aerie',
    'version': '0.7.1',
    'description': 'An async database toolkit.',
    'long_description': '# Aerie\n\nA wrapper around SQLAlchemy made to support asynchronous workloads.\n\n[Aerie](https://baldursgate.fandom.com/wiki/Aerie) - is an [avariel](https://baldursgate.fandom.com/wiki/Elf#Avariel)\n(or winged elf) from Baldur\'s Gate II game.\n\n![PyPI](https://img.shields.io/pypi/v/aerie)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/aerie/Lint)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/aerie)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/aerie)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/aerie)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/aerie)\n![Lines of code](https://img.shields.io/tokei/lines/github/alex-oleshkevich/aerie)\n\n## Installation\n\nInstall `aerie` using PIP or poetry:\n\n```bash\npip install aerie[postgresql]\n# or\npoetry add aerie[postgresql]\n```\n\nFor SQLite use "sqlite" extra. To install all drivers use "full" extra.\n\n## Features\n\n- full async/await support\n- plain SQL with bound params\n- SQLAlchemy query builders support\n- SQLAlchemy ORM support\n- pagination\n\n## TODO\n\n* simplify column definition: `sa.Column(sa.Integer)` -> `models.IntergerField()`\n* integrate with Alembic CLI\n\n## Quick start\n\nSee example application in [examples/](examples/) directory of this repository.\n\n## Usage\n\nAs this library is based on [SQLAlchemy](https://docs.sqlalchemy.org/en/14/index.html), it is strictly recommended\ngetting yourself familiar with it.\n\nA general usage is:\n\n* create an instance of Aerie\n* define ORM models\n* create tables in the database (or, preferably, use Alembic migrations)\n* obtain a session and perform database queries\n\n### Aerie instance\n\nCreate an instance of Aerie class and pass a connection string to it:\n\n```python\nfrom aerie import Aerie\n\ndb = Aerie(\'sqlite+aiosqlite:///tmp/database.sqlite2\')\n# or\ndb = Aerie(\'postgresql+asyncpg://postgres:postgres@localhost/aerie\')\n```\n\n> You need appropriate driver installed. Add "aiosqlite" for SQLite support, or add "asyncpg" for PostreSQL support.\n\n### Raw SQL queries\n\nAt this step Aerie is ready to work. Create a new transaction and execute any query you need.\n\n```python\nfrom sqlalchemy.sql import text\n\n# assuming "users" table exists\nsql = text(\'select * from users where user_id = :user_id\')\nrows = await db.query(sql, {\'user_id\': 1}).all()\n```\n\nFull listing [examples/raw_sql.py](examples/raw_sql.py)\n\n### Using query builder\n\nSure, you are not limited to plain SQL. SQLAlchemy query builders also supported (because Aerie is a tiny layer on top\nof the SQLAlchemy)\n\n```python\nfrom sqlalchemy.sql import text\nimport sqlalchemy as sa\nfrom aerie import metadata\n\nusers = sa.Table(\n    \'users\', metadata,\n    sa.Column(sa.Integer, name=\'id\', primary_key=True),\n    sa.Column(sa.String, name=\'name\'),\n)\n\n# create tables\nawait db.schema.create_tables()\n\nstmt = select(users).where(users.c.id == 2)\nrows = await db.query(stmt).all()\n```\n\nFull listing [examples/tables.py](examples/tables.py)\n\n### Using ORM models and sessions\n\nAnother option to low-level query builder are ORM models. Aerie provides `aerie.Model` class that you should extend to\ncreate your model.\n\n```python\nfrom aerie import Base\nimport sqlalchemy as sa\n\n\nclass User(Base):\n    __tablename__ = \'users\'\n\n    id = sa.Column(sa.Integer, primary_key=True)\n    name = sa.Column(sa.String)\n\n\n# create tables\nawait db.schema.create_tables()\n\nasync with db.session() as session:\n    session.add_all([\n        User(id=1, name=\'One\'),\n        User(id=2, name=\'Two\'),\n        User(id=3, name=\'Three\'),\n    ])\n    await session.flush()\n\n    # get first user in the row set\n    user = await session.query(User).first()\n```\n\n> Make sure the module with models is imported before you create tables.\n> Otherwise they will not be added to the `metadata` and, thus, not created.\n\nFull listing [examples/orm.py](examples/orm.py)\n\n### Pagination\n\nAerie\'s DbSession ships with pagination utilities out of the box. When you need to paginate a query just\ncall `DbSession.paginate` method.\n\n```python\nasync with db.session() as session:\n    page = await session.query(User).paginate(page=1, page_size=10)\n\n    for user in page:\n        print(user)\n\n    print(\'Next page: %s\' % page.next_page)\n    print(\'Previous page: %s\' % page.previous_page)\n    print(\'Displaying items: %s - %s\' % (page.start_index, page.end_index))\n```\n\nThe page object has more helper attributes:\n\n| Property      | Type | Description                                                                                              |\n|---------------|------|----------------------------------------------------------------------------------------------------------|\n| total_pages   | int  | Total pages in the row set.                                                                              |\n| has_next      | bool | Test if the next page is available.                                                                      |\n| has_previous  | bool | Test if the previous page is available.                                                                  |\n| has_other     | bool | Test if there are another pages except current one.                                                      |\n| next_page     | int  | Next page number. Always returns an integer. If there is no more pages the current page number returned. |\n| previous_page | int  | Previous page number. Always returns an integer. If there is no previous page, the number 1 returned.    |\n| start_index   | int  | The 1-based index of the first item on this page.                                                        |\n| end_index     | int  | The 1-based index of the last item on this page.                                                         |\n| total_rows    | int  | Total rows in result set.                                                                                |\n\n## Alembic migrations\n\nAlembic usage is well documented in the official\ndocs: [Using Asyncio with Alembic](https://alembic.sqlalchemy.org/en/latest/cookbook.html#using-asyncio-with-alembic)\n\nNote, you need to use `aerie.metadata` when you configure `target_metadata` option:\n\n```python\n# migrations/env.py\n\nfrom aerie import metadata\n\ntarget_metadata = metadata\n```\n\nAlso, don\'t forget to import all your models in Alembic\'s `env.py` file so their structure is fully loaded and no models\nforgotten.\n\n## Shared instances\n\nYou can configure Aerie to populate Aerie.instances class-level variable, so you can access database instances from\nanywhere of your code. For that, just pass `name` argument to Aerie constructor.\n\n```python\n# migrations/env.py\n\nfrom aerie import Aerie\n\ndb = Aerie(name=\'shared\', ...)\n\n# other file\ndb = Aerie.get_instance(\'shared\')\n```\n\n> Note, instances without name cannot be shared.\n',
    'author': 'Alex Oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/aerie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)



import typing as t
from sqlalchemy import text
from sqlalchemy.engine import Result, Row
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.sql import Executable

from aerie.collections import Collection
from aerie.utils import convert_exceptions


class ResultProxy:
    def __init__(self, engine: AsyncEngine, stmt: t.Union[str, Executable], params: t.Mapping = None) -> None:
        self._engine = engine
        self._stmt = text(stmt) if isinstance(stmt, str) else stmt
        self._params = params

    async def all(self) -> Collection[Row]:
        result = await self._execute()
        return Collection(result.all())

    async def one(self) -> Row:
        with convert_exceptions():
            result = await self._execute()
            return result.one()

    async def one_or_none(self) -> t.Optional[Row]:
        with convert_exceptions():
            result = await self._execute()
            return result.one_or_none()

    async def first(self) -> t.Optional[Row]:
        result = await self._execute()
        return result.first()

    async def scalar(self) -> t.Any:
        with convert_exceptions():
            result = await self._execute()
            return result.scalar()

    async def scalars(self) -> Collection[t.Any]:
        result = await self._execute()
        return Collection(result.scalars())

    async def scalar_one_or_none(self) -> t.Optional[t.Any]:
        with convert_exceptions():
            result = await self._execute()
            return result.scalar_one_or_none()

    async def unique(self, strategy: t.Callable = None) -> t.Optional[t.Any]:
        result = await self._execute()
        return result.unique(strategy)

    async def partitions(self, size: int = None) -> t.Iterator[t.List[t.Any]]:
        result = await self._execute()
        return result.partitions(size)

    async def _execute(self) -> Result:
        async with self._engine.begin() as connection:
            return await connection.execute(self._stmt, self._params)

    def __await__(self) -> t.Generator[Result, None, Result]:
        return self._execute().__await__()
